from threading import Thread
from time import sleep

class EventListener():
  def __init__(self, event, handler, poll_interval, opts={}):
    self.filter = event.createFilter(fromBlock=1)
    self.poll_interval = poll_interval
    self.handler = handler

  def loop(self):
    while True:
      for event in self.filter.get_new_entries():
        self.handler(event)
      sleep(self.poll_interval)

  def run(self):
    self.worker = Thread(target=self.loop)
    self.worker.start()

  def stop(self):
    self.worker.join()


import json
import os
from random import randint
import ipaddress
from time import sleep


def get_receipt(skale, tx):
    return skale.web3.eth.getTransactionReceipt(tx)


def await_receipt(skale, tx, retries=10, timeout=5):
    for _ in range(0, retries):
        receipt = get_receipt(skale, tx)
        if (receipt != None):
            return receipt
        sleep(timeout)
    return None


def check_port(port):
    if port not in range(1, 65535):
        raise ValueError(f'{port} does not appear to be a valid port. Allowed range: 1-65535')


def check_ip(ip):
    return ipaddress.ip_address(ip)


def generate_ws_addr(ip, port):
    check_ip(ip)
    check_port(port)
    return 'ws://' + ip + ':' + str(port)


def get_default_abipath():
    current_dir = os.path.dirname(os.path.realpath(__file__))
    abi_dir = os.path.join(current_dir, os.pardir)
    return os.path.join(abi_dir, 'contracts_data.json')


def get_abi(abi_filepath=None):
    if not abi_filepath:
        abi_filepath = get_default_abipath()
    with open(abi_filepath) as data_file:
        data = json.load(data_file)
    return data


def generate_nonce():
    return randint(0, 65534)



from web3 import Web3


class BaseContract:
    def __init__(self, skale, name, address, abi):
        self.skale = skale
        self.web3 = skale.web3
        self.name = name
        self.address = Web3.toChecksumAddress(address)
        self.abi = abi
        self.contract = self.web3.eth.contract(address=self.address, abi=self.abi)



from skale.contracts import BaseContract
import skale.utils.helper as Helper

class Token(BaseContract):
  def __init__(self, skale):
    name = 'token'
    super().__init__(skale, name, skale.abi[f'{name}_address'], skale.abi[f'{name}_abi'] )



import skale.utils.helper as Helper
from web3 import Web3, WebsocketProvider
import skale.contracts as contracts

class Skale:
  def __init__(self, ip, port, abi_filepath=None):
    ws_addr = Helper.generate_ws_addr(ip, port)
    self.web3 = Web3(WebsocketProvider(ws_addr))
    self.abi = Helper.get_abi(abi_filepath)
    self.__contracts = {}
    self.__init_contracts()

  def __init_contracts(self):
    self.add_contract('manager', contracts.NodeManager(self))
    self.add_contract('token', contracts.Token(self))

  def add_contract(self, name, skale_contract):
    self.__contracts[name] = skale_contract

  def get_contract_by_name(self, name):
    return self.__contracts[name]

  def manager(self):
    return self.get_contract_by_name('manager')

  def token(self):
    return self.get_contract_by_name('token')





import sys

if sys.version_info < (3, 5):
    raise EnvironmentError("Python 3.5 or above is required")

from skale.main import Skale
from skale.contracts import NodeManager, Token, BaseContract
from skale.event_listener import EventListener



"""
Based on https://github.com/pypa/sampleproject
See https://github.com/pypa/sampleproject/blob/master/setup.py for explanations
"""

from setuptools import setup, find_packages

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

setup(
  name='skale-py',
  version='0.2.2',
  description='Skale client tools',

  long_description=long_description,
  long_description_content_type='text/markdown',

  url='https://github.com/GalacticExchange/blockchain-node',
  author='Skale Labs',
  author_email='support@skalelabs.com',

  classifiers=[
    'Development Status :: 2 - Pre-Alpha',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Programming Language :: Python :: 3.6',
  ],

  packages=find_packages(exclude=['contrib', 'docs', 'tests']),
  install_requires=[
    'peppercorn',
    'web3',
    'asyncio',
    'populus'
  ],

  include_package_data=True,
  package_data={  # Optional
    'contracts': ['utils/contracts_data.json'],
  },
)



from web3 import Web3
import socket

from skale.contracts import BaseContract
import skale.utils.helper as Helper
from skale.utils.constants import NODE_DEPOSIT, GAS, OP_TYPES


class NodeManager(BaseContract):
    def __init__(self, skale):
        name = 'node_manager'
        super().__init__(skale, name, skale.abi[f'{name}_address'], skale.abi[f'{name}_abi'])

    def create_node(self, ip, port, wallet):
        token = self.skale.get_contract_by_name('token')

        transaction_data = self.create_node_data_to_bytes(ip, port, wallet['address'])

        nonce = self.skale.web3.eth.getTransactionCount(wallet['address'])
        op = token.contract.functions.transfer(self.address, NODE_DEPOSIT, transaction_data)
        create_node_txn = op.buildTransaction({
            'gas': GAS['create_node'],
            'nonce': nonce
        })

        self.skale.web3.eth.enable_unaudited_features() # todo: deal with this!
        signed_txn = self.skale.web3.eth.account.signTransaction(create_node_txn, private_key=wallet['private_key'])
        return self.skale.web3.eth.sendRawTransaction(signed_txn.rawTransaction)

    def create_node_data_to_bytes(self, ip, port, address):
        address_fx = Web3.toChecksumAddress(address)[2:]  # cut 0x
        nonce = Helper.generate_nonce()

        type_bytes = OP_TYPES['create_node'].to_bytes(1, byteorder='big')
        port_bytes = port.to_bytes(4, byteorder='big')
        nonce_bytes = nonce.to_bytes(4, byteorder='big')  # todo
        ip_bytes = socket.inet_aton(ip)
        address_bytes = bytes.fromhex(address_fx)

        return type_bytes + port_bytes + nonce_bytes + ip_bytes + address_bytes

    def get_bounty(self, node_id, account):
        transaction_opts = {
            'from': account,
            'gas': GAS['get_bounty']
        }
        return self.contract.functions.getBounty(node_id).transact(transaction_opts)

    def send_verdict(self, validator, node_id, downtime, latency, account):
        transaction_opts = {
            'from': account,
            'gas': GAS['send_verdict']
        }
        return self.contract.functions.sendVerdict(validator, node_id, downtime, latency).transact(transaction_opts)

    def get_validated_array(self, node_id, account):
        return self.contract.functions.getValidatedArray(node_id).call({'from': account})

    def __get_node_raw(self, node_id):
        return self.contract.functions.getNode(node_id).call()

    def get_node(self, node_id):
        node_arr = self.__get_node_raw(node_id)
        return self.node_arr_to_obj(node_arr)

    def node_arr_to_obj(self, node_arr):
        return {
            'ip': socket.inet_ntoa(node_arr[0]),
            'port': node_arr[1],
            'status': node_arr[2],
            'public_key': node_arr[3],
            'last_reward_date': node_arr[4],
            'leaving_date': node_arr[5],
            'start_date': node_arr[6]
        }

    def get_active_node_ids(self):
        return self.contract.functions.getActiveNodeIds().call()